package truebar.example;

import truebar.client.AudioRecorder;
import truebar.client.TruebarAuthenticator;
import truebar.client.TruebarClient;
import truebar.client.misc.Pipeline;
import truebar.client.model.ServiceParameters;
import truebar.client.model.pipeline.StageIn;
import truebar.client.model.pipeline.Task;
import truebar.client.model.pipeline.TextSegmentDto;
import truebar.client.model.pipeline.TokenDto;
import truebar.client.model.pipeline.config.StageConfigIn;

import javax.sound.sampled.*;
import java.io.IOException;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/* *****************************************************
    !IMPORTANT!
    This file is just for demonstration purposes
    truebar.client.TruebarClient can be used asynchronously without blocking main thread instead of waiting for responses like it is done in this example.
    AudioRecording can also be performed in separate thread to minimize voice distortions (such as : skipped samples, jitter, etc).
 ******************************************************** */

public class Main {

    public static void main(String[] args) throws ExecutionException, InterruptedException, TimeoutException {
        try {
            Scanner scanner = new Scanner(System.in);

            // ************************************
            // Authentication
            // ************************************
            System.out.print("Enter username: ");
            String userName = scanner.nextLine();
            System.out.print("Enter password: ");
            String password = scanner.nextLine();

            // Create truebar authenticator used to manage authentication details
            TruebarAuthenticator truebarAuthenticator = new TruebarAuthenticator(
                    new ServiceParameters(
                            true,
                            "dev-pipelines-auth.true-bar.si",
                            443,
                            3000
                    ),
                    userName,
                    password
            );

            // Create truebar client instance
            TruebarClient truebarClient = new TruebarClient(
                    truebarAuthenticator,
                    new ServiceParameters(
                            true,
                            "dev-api-pipelines.true-bar.si",
                            443,
                            180000
                    )
            );


            // ************************************
            // Create pipeline using 2 stages: "automatic speech recognition" and "automatic punctuation"
            // Each stage has multiple config options and each config option provides different set of parameters.
            //************************************

            // List all config option provided by ASR stage and print selection menu
            StageIn asrStageIn = truebarClient.getStage(Task.ASR).get();
            List<StageConfigIn> asrStageOptions = asrStageIn.configOptions();
            System.out.println("Available ASR options:");
            for (int i = 0; i < asrStageOptions.size(); i++) {
                System.out.printf("%d -> %s%n", i + 1, asrStageOptions.get(i).tag());
            }
            System.out.print("Enter selection: ");
            StageConfigIn selectedAsrConfigOption = asrStageOptions.get(Integer.parseInt(scanner.nextLine()) - 1);

            // List all config option provided by PC stage and print selection menu
            StageIn pcStageIn = truebarClient.getStage(Task.NLP_pc).get();
            List<StageConfigIn> pcStageOptions = pcStageIn.configOptions();
            System.out.println("Available punctuation options:");
            System.out.printf("%d -> %s%n", 0, "None");
            for (int i = 0; i < pcStageOptions.size(); i++) {
                System.out.printf("%d -> %s%n", i + 1, pcStageOptions.get(i).tag());
            }
            System.out.print("Enter selection: ");
            StageConfigIn selectedPcConfigOption = null;
            int selectedPcIndex = Integer.parseInt(scanner.nextLine());
            if (0 < selectedPcIndex) {
                selectedPcConfigOption = pcStageOptions.get(selectedPcIndex-1);
            }

            // Create pipeline with ASR -> PC stages
            // Enable interim transcripts on ASR stage if supported by selected config option
            Map<String, Object> asrStageParameters = new HashMap<>();
            asrStageParameters.put("enableUnks", false);
            if (selectedAsrConfigOption.parameters().containsKey("enableInterims")) {
                asrStageParameters.put("enableInterims", true);
            }
            if (selectedAsrConfigOption.parameters().containsKey("enableDiarization")) {
                asrStageParameters.put("enableDiarization", true);
            }
            Pipeline pipeline = new Pipeline();
            pipeline.addStage(Task.ASR, selectedAsrConfigOption.tag(), asrStageParameters);
            if (selectedPcConfigOption != null) {
                pipeline.addStage(Task.NLP_pc, selectedPcConfigOption.tag(), Map.of("enableSplitToSentences", false));
            }

            System.out.print("Options:\n1 -> Start streaming session\n2 -> Process file\nEnter selection: ");
            switch (scanner.nextLine()) {
                case "1": {
                    System.out.println("Creating session...");
                    TruebarClient.StreamingSession streamingSession = truebarClient.process(
                                    null,
                                    pipeline,
                                    (textSegment) -> System.out.println(createHypothesis(textSegment)),
                                    (warning) -> System.err.printf("Received following warning from truebar service: %s%n", warning),
                                    (error) -> System.err.printf("Received following error from truebar service: %s%n", error.message())
                            )
                            .get(300, TimeUnit.SECONDS);

                    System.out.println("Available audio inputs:");
                    List<Mixer.Info> audioSources = AudioRecorder.findAvailableAudioMixers();
                    for (int i = 0; i < audioSources.size(); i++) {
                        System.out.printf("%d -> %s%n", i + 1, audioSources.get(i).getName());
                    }
                    Mixer.Info audioInputInfo = audioSources.get(Integer.parseInt(scanner.nextLine()) - 1);

                    try (AudioRecorder audioRecorder = new AudioRecorder(audioInputInfo)) {
                        System.out.println("...CAPTURING AUDIO .....TO FINISH PRESS ENTER...");
                        audioRecorder.start((audioChunk) -> {
                            streamingSession.sendData(audioChunk.array());
                        });

                        // Loop until user presses enter
                        while (System.in.available() == 0) {
                            Thread.sleep(100);
                        }

                        System.out.println("FINISHING SESSION");
                        streamingSession.finish(false).get(3, TimeUnit.SECONDS);
                    }
                    break;
                }
                case "2": {
                    System.out.print("Enter path to audio file: ");
                    String audioFilePath = scanner.nextLine();

                    System.out.println("Uploading file...");
                    List<TextSegmentDto> result = truebarClient
                            .process(pipeline, Path.of(audioFilePath))
                            .get();

                    System.out.println("Result:");
                    for (TextSegmentDto textSegment : result) {
                        System.out.println(createHypothesis(textSegment));
                    }
                    break;
                }
                default: {
                    throw new RuntimeException("Invalid selection!");
                }
            }
        } catch (Exception e) {
            System.out.println("Caught unhandled exception.");
            e.printStackTrace();
        }

        System.out.println("ALL OK");
    }

    private static String createHypothesis(TextSegmentDto textSegment) {
        StringBuilder hypothesis = new StringBuilder();
        boolean previousRightHanded = false;
        for (TokenDto tokenDto : textSegment.tokens()) {
            if (!previousRightHanded && !tokenDto.isLeftHanded()) {
                hypothesis.append(" ");
            }
            hypothesis.append(tokenDto.text());
            previousRightHanded = tokenDto.isRightHanded();
        }
        return hypothesis.toString();
    }
}